#!/bin/sh
# common.sh / 2021.09.12

# -----------------------------------------------------------------------------------#
# AvrLogger - Monitoring and visualization tool.                                     #
# Copyright (C) 2018 by BigRonin.                                                    #
# Member of the German Synology Community Forum "https://www.synology-forum.de"      #
# -----------------------------------------------------------------------------------#
# This file is part of "AvrLogger.spk"                                               #
#                                                                                    #
# AvrLogger is free software: you can redistribute it and/or modify                  #
# it under the terms of the GNU General Public License as published by               #
# the Free Software Foundation, either version 3 of the License, or                  #
# (at your option) any later version.                                                #
# -----------------------------------------------------------------------------------#

dsmVersion="/etc.defaults/VERSION"

# -----------------------------------------------------------------------------------
# --- Check DSM major number ...
# -----------------------------------------------------------------------------------
if [[ "$(grep -m 1 "major" "${dsmVersion}" | sed 's/^[^"]*"\([^"]*\)".*/\1/g')" -lt 7 ]]
   then       DSM6=
   else unset DSM6
fi

# -----------------------------------------------------------------------------------
# --- Check whether we have root rights ...
# -----------------------------------------------------------------------------------
if [[ $EUID -ne 0 ]]
   then unset Root      # No we have not
   else       Root=     # Yes we have :)
fi

# -----------------------------------------------------------------------------------
# --- Check whether we have new FHS (Package Filesystem Hierarchy Standard) ...
# -----------------------------------------------------------------------------------
for iID in $( echo /volume*/@appconf ); do
   if [[ "${iID:7:1}" != "*" ]]; then
      if [ -v appConf ]
         then appConf+=( "$iID" )
         else appConf=(  "$iID" )
      fi
   fi
done

# -----------------------------------------------------------------------------------
if [[ -v SYNOPKG_PKGNAME && ! -v SYNOPKG_REPLICATION && $# -eq 0 ]]; then

   # --------------------------------------------------------------------------------
   # --- Called via spk manager ...
   GUI="/usr/syno/synoman/webman/3rdparty/${SYNOPKG_PKGNAME}"
   ETC="/var/packages/$SYNOPKG_PKGNAME/etc"
   Target="${SYNOPKG_PKGDEST}/ui"

   # --------------------------------------------------------------------------------
   # --- Init Log- / Error-Files ...
   # --------------------------------------------------------------------------------
   ERROR="${ETC}/log/error"
   LOG=""
   ERR=""

   if [ -v appConf ]
      then logPath="${SYNOPKG_PKGDEST_VOL}/@appconf/${SYNOPKG_PKGNAME}" # DSM >= 7.0-41330
      else logPath="/usr/syno/etc/packages/${SYNOPKG_PKGNAME}"          # DSM <  7.0-41330
   fi

   if [[ -v Root && ! -v appConf && ! -d "${logPath}" ]]; then
      mkdir -p "${logPath}"
   fi

   if [[ -d "${logPath}" ]]
   then
      logOK=
      if [ -v appConf ]; then
         LOG="${logPath}/tmp.log"
         ERR="${logPath}/tmp_err.log"
      else
         LOG="${logPath}/${SYNOPKG_PKGNAME}.log"
         ERR="${logPath}/${SYNOPKG_PKGNAME}_err.log"
      fi

      touch "$LOG"
      touch "$ERR"

      if [ ! -v DSM6 ]; then
         chown "AvrLogger:AvrLogger" "${LOG}"
         chown "AvrLogger:AvrLogger" "${ERR}"
      fi

      exec 1>>"$LOG"
      exec 2>>"$ERR"
   else
      unset logOK
   fi

   # --------------------------------------------------------------------------------
   # --- Integrate libraries ...
   . "$(dirname "$0")/asset.sh"

   # --------------------------------------------------------------------------------
   # --- Exit script if try to use an uninitialised variable
   set -o nounset

   # --------------------------------------------------------------------------------
   # --- General initialization part ...
   logConf="$ETC/config"
fi

# -----------------------------------------------------------------------------------
Time="$(date '+%Y.%m.%d %H:%M:%S')"

# -----------------------------------------------------------------------------------
# preinst: This script is run before the package files are transferred to @appstore.
# -----------------------------------------------------------------------------------
pkg_preinst() {
   exit 0
}

# -----------------------------------------------------------------------------------
# postinst: This script is run after the package files are transferred to @appstore.
# -----------------------------------------------------------------------------------
pkg_postinst() {

   unset         Inst
   declare -g -A Inst
   Inst["Version"]="$SYNOPKG_PKGVER"
   Inst["sysHostName"]="$(<${sysHostName})"
   Inst["DsmLanguage"]="${SYNOPKG_DSM_LANGUAGE}"

   tmpConf="${Target}/templates/etc"

   # --------------------------------------------------------------------------------
   echo "---------------------------------------------------------------------------------"
   echo "${Time} PostInst  : ${SYNOPKG_PKG_STATUS} v${SYNOPKG_PKGVER}"
   echo "---------------------------------------------------------------------------------"

   if [ -d "${Target}/helptoc" ]; then
       mv "${Target}/helptoc" "${Target}/.helptoc"
      echo "${Time} PostInst  : Renamed \"/helptoc\" to \"/.helptoc\""
   fi

   # --------------------------------------------------------------------------------
   # --- Perform an "AvrLogger" installation ...
   # --------------------------------------------------------------------------------
   if [ "$SYNOPKG_PKG_STATUS" == "INSTALL" ]; then
      # -----------------------------------------------------------------------------
      # --- First : Copy "etc" directory structure ...
      cp -r "$tmpConf/." "$ETC/"
      echo "${Time} PostInst  : Directory structure \"etc\" created"

      # -----------------------------------------------------------------------------
      pkg_Virtual    # --- Determine if the DSM is running in a virtual environment ...
      pkg_FanStatus  # --- Determine whether Fan-Control is supported on this NAS ...
      TimeOut=$( pkg_CpuScore )

      if [ -v Virtual ]; then
         echo "${SYNOPKG_PKGNAME}: Syno-Model is \"${Virtual}\". The functionality is limited." > $SYNOPKG_TEMP_LOGFILE
         Inst["sysVirtualDSM"]="Yes"
      else
         Inst["sysVirtualDSM"]="No"
      fi

      if [ -v ValidFanMode ]
         then Inst["sysExcludeSCEMD"]="No"
         else Inst["sysExcludeSCEMD"]="Yes"
      fi

      # -----------------------------------------------------------------------------
      Inst["SpkVolume"]="${SYNOPKG_PKGDEST_VOL}"
      Inst["sysCpuScore"]="${TimeOut}"
      pkg_UpdateConfig "${tmpConf}/config/log.conf" "${ETC}/config/log.conf"
      rm -f "${ETC}/config/log.conf.save"

   fi
   # ----------------------- END of INSTALL -----------------------------------------
   # --------------------------------------------------------------------------------

   # --------------------------------------------------------------------------------
   # --- Perform an "AvrLogger" update ...
   # --------------------------------------------------------------------------------
   if [ "$SYNOPKG_PKG_STATUS" == "UPGRADE" ]; then
      # --- Copy default Mail-Config ...
      cp "${tmpConf}/config/mail.conf" "${ETC}/config/mail.conf"

      # -----------------------------------------------------------------------------
      pkg_Virtual    # --- Determine if the DSM is running in a virtual environment ...
      pkg_FanStatus  # --- Determine whether Fan-Control is supported on this NAS ...
      TimeOut=$( pkg_CpuScore )

      if [ -v Virtual ]
         then Inst["sysVirtualDSM"]="Yes"
         else Inst["sysVirtualDSM"]="No"
      fi

      if [ -v ValidFanMode ]
         then Inst["sysExcludeSCEMD"]="No"
         else Inst["sysExcludeSCEMD"]="Yes"
      fi

      # -----------------------------------------------------------------------------
      # --- Update the AvrLogger Config (log.conf) ...
      Inst["sysCpuScore"]="${TimeOut}"
      pkg_UpdateConfig "${tmpConf}/config/log.conf" "${ETC}/config/log.conf"
      rm -f "${ETC}/config/log.conf.save"
      echo "${Time} PostInst  : Config \"log.conf\" updated"

      # -----------------------------------------------------------------------------
      touch "$UpgradePKG"

   fi
   # ----------------------- END of UPGRADE -----------------------------------------
   # --------------------------------------------------------------------------------

   # --------------------------------------------------------------------------------
   # --- Set the Shared Memory "ExcludeSHM" flag according to the user input ...
   if [[ "$wizCheckShm" == "true" ]]
   then  # --- Enable use of "Shared Memory" ...
      rm -f "$ExcludeSHM"
      echo "${Time} PostInst  : Use of \"Shared Memory\" is enabled"
   else  # --- Disable use of "Shared Memory" ...
      touch "$ExcludeSHM"
      echo "${Time} PostInst  : Use of \"Shared Memory\" is disabled"
   fi

   # --------------------------------------------------------------------------------
   # --- Create "Default-Template" ...
   cp "${tmpConf}/config/log.conf" "${Target}/templates/Layout/Default.cnf"
   pkg_UpdateConfig "${tmpConf}/config/log.conf" "${Target}/templates/Layout/Default.cnf"
   echo "${Time} PostInst  : Config \"Default.cnf\" created"

   # --------------------------------------------------------------------------------
   echo "${Time} PostInst  : Global TimeOut set to \"${TimeOut}\""
   echo "${TimeOut}" > "${cpuScore}"

   # --------------------------------------------------------------------------------
   if [ -v DSM6 ]; then
      [ -L "${GUI}" ] && rm -f "${GUI}"
      ln -s "${SYNOPKG_PKGDEST}/ui" "${GUI}"
      echo "${Time} PostInst  : Symbolic link in 3rdparty created"
   fi

   if [ -v logOK ]; then
      if [ -s "$ERR" ]
      then
         echo "-- ERROR Log : ------------------------------------------------------------------"
         cat "$ERR"
         echo "---------------------------------------------------------------------------------"
      else
         echo "---------------------------------------------------------------------------------"
         echo "${Time} ${SYNOPKG_PKGNAME} v${SYNOPKG_PKGVER} ${SYNOPKG_PKG_STATUS} successful."
         echo "---------------------------------------------------------------------------------"
      fi

      cat "$LOG" >> "${ERROR}"

      > "$LOG"
      > "$ERR"
   fi

   touch "${NewsAvrPKG}"
   rm -f "${SynoCheck}"

   exit 0
}

# -----------------------------------------------------------------------------------
# postuninst: This script is run before the package is removed.
# -----------------------------------------------------------------------------------
pkg_preuninst() {

   echo "${Time} PreUninst : ${SYNOPKG_PKG_STATUS} is being prepared ..."

   if [ "$SYNOPKG_PKG_STATUS" == "UNINSTALL" ]; then

      if [ -v DSM6 ]; then
         rm -f "${GUI}"
         echo "${Time} PreUninst : Symbolic link in 3rdparty deleted"
      fi

      pkg_Virtual
      pkg_FanStatus

      # -------------------------------------------------------------------------------
      # --- Check Files & Hashes ...
      md5_Conf="$(  get_key_value "${ETC}/config/log.conf" 'FanScemdHash' )"
      [ -f "${ETC}/log/scemd.xml" ] && md5_Scemd="$( md5sum "${ETC}/log/scemd.xml" | awk '{print $1}' )"
      [ -f "${ETC}/log/scemd.md5" ] && md5_Hash="$(<"${ETC}/log/scemd.md5")"

      if [[ "$wizCheckScemd" == "true" && ! -v Virtual && ! -f "$ExcludeSCEMD" ]]; then

         if [ -v md5_Scemd ] && [ -v md5_Hash ] && [ "${md5_Hash}" == "${md5_Scemd}" ] && [ "${md5_Hash}" == "${md5_Conf}" ]; then
            # -------------------------------------------------------------------------------
            # --- Restore "scemd.xml" ...
            cp "${ETC}/log/scemd.xml" "/usr/syno/etc.defaults/scemd.xml"
            synoservicectl --restart scemd > /dev/null 2>&1 & 
            echo "${Time} PreUninst : The \"scemd.xml\" was successfully restored"
         fi
      fi

      if [[ -n "$wizUserDir" ]]; then
         eval arr=("$(pkg_FindeShare "$wizUserDir")")
         if [ -v arr ]; then
            ExportPath="${arr[1]}/${arr[0]}"
         fi
      fi

      if [[ -v ExportPath && "$wizSaveData" == "true" ]]; then
         echo "${Time} PreUninst : Secure (save) Data in : \"${ExportPath}\""

         if [ -f "${ETC}/tmp/AvrLogger.dat" ]; then
            cp   "${ETC}/tmp/AvrLogger.dat" "${ExportPath}/AvrLogger.dat"
            echo "${Time} PreUninst : Backed up \"AvrLogger.dat\""
         fi

         if [ -f "${ETC}/log/error" ]; then
            cp   "${ETC}/log/error" "${ExportPath}/error"
            echo "${Time} PreUninst : Backed up \"error\""
         fi

         if [ -f "${ETC}/log/enumDrives" ]; then
            cp   "${ETC}/log/enumDrives" "${ExportPath}/enumDrives"
            echo "${Time} PreUninst : Backed up \"enumDrives\""
         fi

         if [ -f "${ETC}/log/enumStorage" ]; then
            cp   "${ETC}/log/enumStorage" "${ExportPath}/enumStorage"
            echo "${Time} PreUninst : Backed up \"enumStorage\""
         fi

         if [ -f "${ETC}/config/log.conf" ]; then
            cp   "${ETC}/config/log.conf" "${ExportPath}/log.conf"
            echo "${Time} PreUninst : Backed up \"log.conf\""
         fi

         while IFS= read -r dev; do
            [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue
            cp "${dev}" "${ExportPath}/tmp.svg"
            echo "${Time} PreUninst : Backed up \"tmp.svg\""
            break
         done <<< "$( find  "${Target}/images/" -maxdepth 1 -name "tmp_*.svg" )"

         cat "$LOG" >> "${ExportPath}/error"
         [ -s "$ERR" ] && cat "$ERR" >> "${ExportPath}/error"
         chown -R admin:users "${ExportPath}"
         chmod -R 755 "${ExportPath}"
      fi
   fi

   exit 0
}

# -----------------------------------------------------------------------------------
# postuninst: This script is run after the package is removed.
# -----------------------------------------------------------------------------------
pkg_postuninst() {

   if [ "$SYNOPKG_PKG_STATUS" == "UNINSTALL" ]; then
      [ ! -v appConf ] && rm -rf "/usr/syno/etc/packages/${SYNOPKG_PKGNAME}"
      rm -f  "/var/log/packages/${SYNOPKG_PKGNAME}.log"
      # -----------------------------------------------------------------------------
      if [ -v DSM6 ]
      then
         rm -f  "/usr/bin/ppython"
         rm -f  "${GUI}"
      else
         [ -v appConf ] && rm -rf "${SYNOPKG_PKGDEST_VOL}/@appconf/${SYNOPKG_PKGNAME}"
         rm -rf "${SYNOPKG_PKGDEST_VOL}/@appdata/${SYNOPKG_PKGNAME}"
         rm -rf "${SYNOPKG_PKGDEST_VOL}/@apphome/${SYNOPKG_PKGNAME}"
         rm -rf "${SYNOPKG_PKGDEST_VOL}/@apptemp/${SYNOPKG_PKGNAME}"
      fi
   fi

   exit 0
}

# -----------------------------------------------------------------------------------
# preupgrade: Package Center will call this script before uninstalling the old version of your package.
# -----------------------------------------------------------------------------------
pkg_preupgrade() {

   echo "---------------------------------------------------------------------------------"
   echo "${Time} PreUpgrade: from v${SYNOPKG_OLD_PKGVER} ..."

   # --------------------------------------------------------------------------------
   # --- Stop all instances of GUI-cgi (AvrLogger.cgi) ...
   if [ $(ps ax | grep -v grep | grep "${SYNOPKG_PKGNAME}.cgi" | awk '{print $1}') ]; then
      echo "${Time} PreUpgrade: Going to stop \"${SYNOPKG_PKGNAME}.cgi\""
      killall -w "${SYNOPKG_PKGNAME}.cgi"
      sleep 0.250
   fi

   # --------------------------------------------------------------------------------
   # --- Save all SVG Preview-Files ...
   # --------------------------------------------------------------------------------
   # --- Remove old Preview-Files ...
   while IFS= read -r dev; do
      [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue
      rm -f "${dev}"
   done <<< "$( find  "${ETC}/" -maxdepth 3 -name "???_*.svg" )"

   # --------------------------------------------------------------------------------
   # --- Copy Preview-Files ...
   while IFS= read -r dev; do
      [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue
      cp "${dev}" "${ETC}/tmp/${dev##*/}"
      [ ! -v DSM6 ] && chown -R "AvrLogger:AvrLogger" "${ETC}/tmp/${dev##*/}"
   done <<< "$( find  "${SYNOPKG_PKGDEST}/" -maxdepth 3 -name "???_*.svg" )"

   # --------------------------------------------------------------------------------
   exit 0
}

# -----------------------------------------------------------------------------------
# postupgrade: Package Center will call this script after installing the new version of your package.
# -----------------------------------------------------------------------------------
pkg_postupgrade() {

   # --------------------------------------------------------------------------------
   # --- Restore SVG Preview-Files ...
   while IFS= read -r dev; do
      [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue
      mv "${dev}" "${Target}/images/${dev##*/}"
   done <<< "$( find  "${ETC}/tmp/" -maxdepth 1 -name "???_*.svg" )"

   > "$LOG"
   > "$ERR"

   rm -f "${MemorySize}"

   exit 0
}

# -----------------------------------------------------------------------------------
# --- Read a Config (cleand) ...
# -----------------------------------------------------------------------------------
# --- $1 = Config-File
# -----------------------------------------------------------------------------------
pkg_ReadConfig() {
   unset         Config
   declare -g -A Config

   local key value

   while IFS= read -r key; do
      value="${key##*=}"
      Config["${key%%=*}"]="${value:1:-1}"
   done <"$1"
}

# -----------------------------------------------------------------------------------
# --- Eine ältere / frühere Konfig-Datei aktualisieren:
# --- 1 : Einlesen der älteren Konfig-Datei (diese muss bereits gesäubert sein.)
# --- 2 : Einlesen der neuen Konfig-Datei.  (diese muss bereits gesäubert sein.)
# --- 3 : Zusammenführen beider Konfig-Dateien- Dabei ncht mehr benötigte Variablen ignorieren.
# --- 4 : Die nun aktualsierte Konfig-Datei, unter beibehaltung der Sortierung, speichern.
# -----------------------------------------------------------------------------------
# --- $1 = Source:      Config-Template
# --- $2 = Destination: previous Config-File (cleand)
# --- $3 = SPK Version
# -----------------------------------------------------------------------------------
pkg_UpdateConfig() {
   
   declare -A    preConf tmpConf
   declare -g -A Config
   local line value key conf

   # --- Read a previous "cleand" config ...
   while IFS= read -r line; do
      value="${line##*=}"
      preConf["${line%%=*}"]="\"${value:1:-1}\""
   done <"$2"
 
   # --- Read the "cleand" Config-Template ...
   while IFS= read -r line; do
      value="${line##*=}"
      tmpConf["${line%%=*}"]="\"${value:1:-1}\""
   done <"$1"

   while IFS= read -r line; do
      # --- Test if "Key" exist in "previous" config ...
      if [[ -v preConf["${line%%=*}"] ]]
         then value="${preConf["${line%%=*}"]}"      # --- Use "previous"-Value
         else value="${line##*=}"                    # --- Use "template"-Value
      fi
      if [[ -v Inst["${line%%=*}"] ]]
         then value="\"${Inst["${line%%=*}"]}\""
      fi
      conf+="${line%%=*}=\"${value:1:-1}\"\n"
      Config["${line%%=*}"]="${value:1:-1}"
   done <"$1"

   # -- Write back updatete config ...
   echo -ne "$conf" > "$2"
}

# -----------------------------------------------------------------------------------
# --- Find "share" : Clean up / correct the path
# -----------------------------------------------------------------------------------
# --- $1 = share-path
# --- >>> Return Array if found:
# -->    arg[0] = Corrected "share" path
# -->    arg[1] = Volume of "share" path
# -----------------------------------------------------------------------------------
pkg_FindeShare() {

   local Volume UserDir=$1

   # --- remove a leading slash "/" ...
   [ "/" == "${UserDir:0:1}" ] && UserDir="${UserDir:1}"
   # --- remove a final slash "/" ...
   [ "/" == "${UserDir: -1}" ] && UserDir="${UserDir:: -1}"
   local tmp="${UserDir,,}"
   # --- remove a volume label at the beginning ...
   [ -z "${tmp%%volume*}" ] && UserDir="${UserDir#*/}"

   # --- Walk through all directories that start with "volume" ...
   for Volume in $(find "/" -maxdepth 1 -name "volume?"); do
      # --- Search for the "UserDir" ...
      if [[ -d "${Volume}/${UserDir}" && -n "$UserDir" ]]; then
         local archive=
         break
      fi
   done

   [ -v archive ] && echo "\"$UserDir\"" "\"$Volume\""
}

# -----------------------------------------------------------------------------------
# --- Check previus Version
# -----------------------------------------------------------------------------------
pkg_Version() {

   oldVersion="$(grep -m 1 "Version" "$1" | sed 's/^[^"]*"\([^"]*\)".*/\1/g')"
   oldVersion=$( echo ${oldVersion} | sed 's/\.//g' | sed 's/\-//g' )
   [ $oldVersion -lt 108017 ] && unset oldVersion  # 1.0.8-017
}

# -----------------------------------------------------------------------------------
# --- Determine whether Fan-Control is supported on this NAS ...
# -----------------------------------------------------------------------------------
pkg_FanStatus() {

   unset Type Speed Mode Pwm
   local line

   while IFS= read -r line; do
      if [ ! -v Type ]; then
         # --------------------------------------------------------------------------
         # --- Finde the first section with "type=" ...
         [[ $line == *"type="* ]] && Type="$line"
      else
         # --------------------------------------------------------------------------
         # --- Find the first "disk_temperature" ...
         if [[ ! -v Speed && $line == *"</disk_temperature>"* ]]; then
            Speed="${line##*fan_speed=\"}"
            Speed="${Speed%%\"*}"
         fi

         # --------------------------------------------------------------------------
         # --- Count the occurrence of '</pwm_config>' ...
         if [[ $line == *"</pwm_config>"* ]]; then
            if [ -v Pwm[@] ]
               then Pwm+=( "$line" )
               else Pwm=(  "$line" )
            fi
         fi

         # --------------------------------------------------------------------------
         # --- Find close Token ...
         if [[ $line == *"</adt_fan_config>"* || $line == *"</fan_config>"* ]]; then
            break
         fi

      fi
   done <"/usr/syno/etc.defaults/scemd.xml"

   # --------------------------------------------------------------------------------
   if [ -v Type ]; then
      if [ -v Pwm ] && [[ $Type == *"<adt_fan_config"* && ${#Pwm[@]} == 1 ]]; then
         Mode="PWM"
      fi

      if [[ $Type == *"<fan_config"* && -v Speed && $Speed == *"hz"* ]]; then
         Mode="ADT"
      fi
   fi

   if [[ -v Mode ]]; then
      # --- Valid Fan-Mode
      rm -f "$ExcludeSCEMD"
      ValidFanMode=
   else
      # --- Unsupported Fan-Mode
      touch "$ExcludeSCEMD"
      unset ValidFanMode
      if [ -v $SYNOPKG_PKG_STATUS ] && [ "$SYNOPKG_PKG_STATUS" == "INSTALL" ]; then
         echo "${Time} PostInst  : Unsupported \"scemd.xml\" - Fan control is disabled."
      fi
   fi
}

# -----------------------------------------------------------------------------------
# --- Determine if the DSM is running in a virtual environment ...
# -----------------------------------------------------------------------------------
pkg_Virtual() {

   Virtual="$(grep -m 1 "upnpmodelname" "/etc.defaults/synoinfo.conf" | sed 's/^[^"]*"\([^"]*\)".*/\1/g')"
   [[ "$Virtual" != "VirtualDSM" && "$Virtual" != "DDSM" ]] && unset Virtual

   if [ -v Virtual ]
   then
      if [ -v SYNOPKG_PKG_STATUS ] && [ "$SYNOPKG_PKG_STATUS" == "INSTALL" ]; then
         echo "${Time} PostInst  : Virtual \"$Virtual\" environment detected."
      fi
      touch "$VirtualDSM"
   else
      rm -f "$VirtualDSM"
   fi
}

# -----------------------------------------------------------------------------------
# CPU performance Score for determining the TimeOut value.
# -----------------------------------------------------------------------------------
pkg_CpuScore() {

   local iFreqScore=0
   local TimeOut=20

   # --------------------------------------------------------------------------------
   # --- Sum up the individual maximum CPU frequencies ...
   if [[ -d /sys/devices/system/cpu && ! -f "${VirtualDSM}" ]]; then
      while IFS= read -r dev; do
         [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue
         iFreqScore=$[ iFreqScore + $( <${dev} ) ]
      done <<< "$( find /sys/devices/system/cpu/cpu*/cpufreq/ -name "cpuinfo_max_freq" )"
   fi

   # --------------------------------------------------------------------------------
   # --- Convert KHz to GHz ...
   iFreqScore=$[ iFreqScore / 1048576 ]

   # --------------------------------------------------------------------------------
   # --- Determine global TimeOut value
   if   [ $iFreqScore -lt 2 ]; then TimeOut=20
   elif [ $iFreqScore -lt 5 ]; then TimeOut=10
   else TimeOut=5
   fi

   echo "${TimeOut}"

}

# -----------------------------------------------------------------------------------
# export: This script is called via Replication service.
# -----------------------------------------------------------------------------------
pkg_export() {

   if [ -d "${ETC}" ]
   then
      if [ -d "${ETC}/config" ]; then
         # -----------------------------------------------------------------------------------
         # --- Secure AvrLogger Config-File "log.conf" ...
         if [ -f "${ETC}/config/log.conf" ]; then
            # -----------------------------------------------------------------------------
            # --- If Daemon is running, wait for Idel or Timeout ...
            # -----------------------------------------------------------------------------
            if [ $(ps ax | grep -v grep | grep "${Name}.py" | awk '{print $1}') ]; then
               rm -f "$Heartbeat"

               local iTimeOut=$( <${cpuScore} )
               local TimeOut=$[ $(date +%s) + iTimeOut ]

               until [[ ! -f "$AvrLock" ]]; do
    	            if [ $(date +%s) -ge $TimeOut ]; then
    	               echo -2
    	               return
                  fi
                  sleep 0.005
                  if [ -f "$Heartbeat" ]; then
                     TimeOut=$[ $(date +%s) + iTimeOut ]
                     rm -f "$Heartbeat"
                  fi
               done
            fi
            # -----------------------------------------------------------------------------------
            cp "${ETC}/config/log.conf" "${ExportPath}/log.conf"
         fi
      fi

      if [  -d "${ETC}/tmp" ]; then
         # -----------------------------------------------------------------------------------
         # --- Secure "AvrLogger.dat" ...
         if [ -f "${ETC}/tmp/AvrLogger.dat" ]; then
            if [ $(ps ax | grep -v grep | grep "${Name}.py" | awk '{print $1}') ]; then
               until [[ ! -f "$FlowFlag" ]]; do
                  sleep 0.005
               done
            fi
            # -----------------------------------------------------------------------------------
            cp "${ETC}/tmp/AvrLogger.dat" "${ExportPath}/AvrLogger.dat"
         fi
      fi
      if [  -d "${ETC}/log" ]; then
         # -----------------------------------------------------------------------------------
         # --- Secure "Hard Disk DataBase" ...
         if [ -f "${ETC}/log/enumDrives" ]; then
            if [ $(ps ax | grep -v grep | grep "${Name}.py" | awk '{print $1}') ]; then
               until [[ ! -f "$FlowFlag" ]]; do
                  sleep 0.005
               done
            fi
            # -----------------------------------------------------------------------------------
            cp "${ETC}/log/enumDrives" "${ExportPath}/enumDrives"
         fi
      fi
   else 
      echo -1
   fi
   echo 0
}

# -----------------------------------------------------------------------------------
# import: This script is called via Replication service.
# -----------------------------------------------------------------------------------
pkg_import() {

   local Err=0

   unset         Inst
   declare -g -A Inst
   Inst["Version"]="$SYNOPKG_PKGVER"

   tmpConf="${SYNOPKG_PKGDEST}/templates/etc"

   # -----------------------------------------------------------------------------------
   # --- Restore AvrLogger Config-File "log.conf" ...
   if [ -f "${ImportPath}/log.conf" ]
   then
      cp "${ImportPath}/log.conf" "${ETC}/config/log.conf"

      # -------------------------------------------------------------------------------
      # --- Update the AvrLogger Config (log.conf) ...
      pkg_UpdateConfig "${tmpConf}/config/log.conf" "${ETC}/config/log.conf"
      rm -f "${ETC}/config/log.conf.save"

      # -----------------------------------------------------------------------------------
      # --- Restore AvrLogger.dat ...
      if [ -f "${ImportPath}/AvrLogger.dat" ]; then
         cp "${ImportPath}/AvrLogger.dat"    "${ETC}/tmp/AvrLogger.dat"
      fi

      # -----------------------------------------------------------------------------------
      # --- Restore "Hard Disk DataBase" ...
      if [ -f "${ImportPath}/enumDrives" ]; then
         cp "${ImportPath}/enumDrives"    "${ETC}/log/enumDrives"
      fi

      echo "${Time} PKG: Data and settings successfully restored, restoration completed."             >> "$ERROR"
      touch "$RestoreConf"
   else
      Err=-1
   fi

   echo $Err
}

# -----------------------------------------------------------------------------------
# --- Prepare "shared memory" ... 
# -----------------------------------------------------------------------------------
pkg_PrepareShm() {
   [ ! -v shm ] && return
   # --------------------------------------------------------------------------------
   pkg_PrepareSub "config"
   pkg_PrepareSub "tmp"

   rm -f "$ShmLock"
}

# -----------------------------------------------------------------------------------
# --- Prepare a single "shared memory"-folder ... 
# -----------------------------------------------------------------------------------
# --- $1 = Folder name
# -----------------------------------------------------------------------------------
pkg_PrepareSub() {
   # --------------------------------------------------------------------------------
   # --- Check the presence of the symbolic link for "folder" ...
   if [ ! -L "/var/packages/${Name}/etc/${1}" ]
   then
      # --- If the symbolic link does not exist ...
      if [ ! -d "${shm}/${Name}/${1}" ]
         then mkdir -p "${shm}/${Name}/${1}"       # --- Ordner erstellen
         else rm -rf   "${shm}/${Name}/${1}/*"     # --- Inhalt des Ordners löschen
      fi

      # --- Init "safe" Foler ...
      if [ ! -d "${ETC}/safe" ]
         then mkdir -p "${ETC}/safe/"${1}                                     # --- Root-"safe"-Ordner erstellen
         else [ ! -d   "${ETC}/safe/${1}" ] && mkdir -p "${ETC}/safe/${1}"    # --- "safe"-Ordner erstellen
      fi

      # --- Copy current Folder into the "safe" ...
      cp -rp "${ETC}/${1}/." "${ETC}/safe/${1}"             # --- Copiere Ordner von ETC nach shm (/dev/shm)

      cp -rp "${ETC}/${1}/."       "${shm}/${Name}/${1}/"   # --- Copiere Ordner von ETC nach shm (/dev/shm)
      mv     "${ETC}/${1}"         "${ETC}/${1}.shm"        # --- Ordner in ETC umbenennen
      ln -s  "${shm}/${Name}/${1}" "${ETC}"                 # --- Symbolischen Link erstellen
      rm -rf                       "${ETC}/${1}.shm"        # --- Umbenannten Ordner löschen
   else
      # --- Wenn der Ordner im Shared Memory vorhanden ist ... alles ok
      [ ! -d "${shm}/${Name}/${1}" ] && pkg_Emergency "${1}"
   fi
}

# -----------------------------------------------------------------------------------
# --- Emergency: Restore from vault ...
# -----------------------------------------------------------------------------------
pkg_Emergency() {

   # meckern !!
   local Time="$(date '+%Y.%m.%d %H:%M:%S')"
   echo "$Time PKG: Fatal error occurred. Unable to access \"${shm}/${Name}/${1}\""    >> "$ERROR"
   echo "$Time PKG: \"$1\" is restored from the vault."                                >> "$ERROR"
   echo "$Time PKG: The data since the last start of the Daemon is probably lost."     >> "$ERROR"

   mkdir -p "${shm}/${Name}/"${1}                        # --- Ordner erstellen
   cp -rp "${ETC}/safe/${1}/." "${shm}/${Name}/${1}"     # --- Daten wiederherstellen
}

# -----------------------------------------------------------------------------------
# --- Dissolve "shared memory" ... 
# -----------------------------------------------------------------------------------
pkg_DissolveShm() {
   [ ! -v shm ] && return
   # --------------------------------------------------------------------------------
   pkg_DissolveSub "config"
   pkg_DissolveSub "tmp"

   # --------------------------------------------------------------------------------
   if [ -d   "${shm}/${Name}" ]; then
      mv     "${shm}/${Name}" "${shm}/${Name}Dissolve"
      rm -rf "${shm}/${Name}Dissolve"
   fi

   rm -f "$ShmLock"
}

# -----------------------------------------------------------------------------------
# --- Dissolve a single "shared memory"-folder ... 
# -----------------------------------------------------------------------------------
# --- $1 = Folder name
# -----------------------------------------------------------------------------------
pkg_DissolveSub() {
   # --- Check if the symbolic link exists ...
   if [ -L "/var/packages/${Name}/etc/${1}" ]; then
      if [ ! -d        "${ETC}/${1}.shm" ]
         then mkdir -p "${ETC}/${1}.shm"
         else rm -rf   "${ETC}/${1}.shm/*"
      fi

      local cp_err="$(cp -rp "${shm}/${Name}/${1}/." "${ETC}/${1}.shm/" 2>&1)"
      rm -f  "/var/packages/${Name}/etc/${1}"

      if [ -n "$cp_err" ]; then
         local Time="$(date '+%Y.%m.%d %H:%M:%S')"
         echo "$Time PKG: Fatal error occurred. Unable to access \"/dev/shm/${1}\""       >> "$ERROR"
         echo "$Time PKG: \"$1\" is restored from the vault."                             >> "$ERROR"
         echo "$Time PKG: The data since the last start of the Daemon is probably lost."  >> "$ERROR"
         cp -rp "${ETC}/safe/${1}/." "${ETC}/${1}.shm/"
      fi

      mv "${ETC}/${1}.shm" "${ETC}/${1}"

   fi
}

# -----------------------------------------------------------------------------------
# --- Get System / User language ...
# -----------------------------------------------------------------------------------
# --- $1 = User Name (admin ?)
# --- $2 = Path inc. File name of/to language Files ( e.g. "/var/packages/wz_lang_"[enu.sh] )
# -----------------------------------------------------------------------------------
pkg_InitLanguage() {

   local tmp lang

   # --------------------------------------------------------------------------------
   # --- Get system language ...
   lang=$(grep -m 1 "language" "/etc/synoinfo.conf" | sed 's/^[^"]*"\([^"]*\)".*/\1/g')
   [[ "$lang" == "def" ]] && lang="enu"

   # --------------------------------------------------------------------------------
   # --- Get user language ...
   if [[ -f "/usr/syno/etc/preference/${1}/usersettings" ]]; then
      local tmp=$(jq -r ".Personal.lang" "/usr/syno/etc/preference/${1}/usersettings")
      [[ -n "$tmp" && "$tmp" != "def" && "$tmp" != "null" ]] && lang=$tmp
   fi

   # --------------------------------------------------------------------------------
   # --- Check if a suitable language file is available ...
   if [[ -f "${2}${lang}.sh" ]]
      then file="${2}${lang}.sh"
      else file="${2}enu.sh"
   fi

   if [ $# -gt 2 ]
      then echo   "$file"
      else source "$file"
   fi
}

# -----------------------------------------------------------------------------------
# --- Change owner to "root" ...
# ->> /var/packages/AvrLogger/scripts/start-stop-status root
# -----------------------------------------------------------------------------------
pkg_ChangeRoot() {

   [ -v DSM6 ] && return

   # --------------------------------------------------------------------------------
   # --- Init some Values ...
   sPackage="/var/packages/${Name}"

   sRoot="root:root"
   sUser="${Name}:${Name}"

   # --------------------------------------------------------------------------------
   # --- Check whether we have new FHS (Package Filesystem Hierarchy Standard)
   # --- and restore the Ownerships if necessary ...
   . "$(dirname "$0")/restore.sh"
   iChange=    # --- Must be set to restore Ownerships
   f_Check     # --- Execute Ckeck
   # --------------------------------------------------------------------------------

   local Time="$(date '+%Y.%m.%d %H:%M:%S')"

   echo                     "Start \"rooting\" ${Name} ..."
   echo "${Time} RootPKG   : Start \"rooting\" ${Name} ..."       >> "$ERROR"

   spkHome=$( echo /volume*/@appstore/${Name}/ui )
   echo                     "APP Home : ${spkHome}"
   echo "${Time} RootPKG   : APP Home : ${spkHome}"               >> "$ERROR"

   if [ -v appConf ]; then
      spkConf="${appConf}/${Name}"
      echo                     "APP Conf : ${spkConf}"
      echo "${Time} RootPKG   : APP Conf : ${spkConf}"            >> "$ERROR"
   fi

   # --------------------------------------------------------------------------------
   # --- Set / exchanged APP-privilege's ...
   # --------------------------------------------------------------------------------
   echo                     "Changeing ownership to : ${sRoot}"
   echo "${Time} RootPKG   : Changeing ownership to : ${sRoot}"   >> "$ERROR"

   while IFS= read -r dev; do
      [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue

      chown "${sRoot}" "${dev}"
      echo                     "Done : chown ${sRoot} ${dev}"
      echo "${Time} RootPKG   : Done : chown ${sRoot} ${dev}"     >> "$ERROR"
   done <<< "$( find  "${spkHome}/" -maxdepth 1 -name "A*" )"

   # --------------------------------------------------------------------------------
   # --- Restore Preview-SVG privilege's ...
   iSVG=

   while IFS= read -r dev; do
      [[ -z "${dev}" ]] || [[ -d "${dev}" ]] && continue

      if [ -v iSVG ]; then
         unset iSVG
         echo                     "Restore SVG ownership to : ${sUser}"
         echo "${Time} RootPKG   : Restore SVG ownership to : ${sUser}"             >> "$ERROR"
      fi

      chown "${sUser}" "${dev}"
      echo                     "Done : chown ${sUser} ${dev}"
      echo "${Time} RootPKG   : Done : chown ${sUser} ${dev}"   >> "$ERROR"
   done <<< "$( find  "${spkHome}/images/" -maxdepth 1 -name "???_*.svg" )"

   # --------------------------------------------------------------------------------
   if [ -f "${spkHome}/templates/root.txt" ]; then
      echo                     "Exchanging the AvrLogger SPK privileges ..."
      echo "${Time} RootPKG   : Exchanging the AvrLogger SPK privileges ..."        >> "$ERROR"

      mv    "${spkHome}/templates/root.txt" "${sPackage}/conf/privilege"
      echo                     "Move : ${spkHome}/templates/root.txt"
      echo "${Time} RootPKG   : Move : ${spkHome}/templates/root.txt"               >> "$ERROR"

      echo                     "To   : ${sPackage}/conf/privilege"
      echo "${Time} RootPKG   : To   : ${sPackage}/conf/privilege"                  >> "$ERROR"

      echo                     "Changeing ownership to : ${sRoot}"
      echo "${Time} RootPKG   : Changeing ownership to : ${sRoot}"                  >> "$ERROR"
      chown "${sUser}" "${sPackage}/conf/privilege"

      echo                     "Done : chown ${sRoot} ${sPackage}/conf/privilege"
      echo "${Time} RootPKG   : Done : chown ${sRoot} ${sPackage}/conf/privilege"   >> "$ERROR"
   fi

   # --------------------------------------------------------------------------------
   # --- This is a precaution to avoid problems with the privilege's ...
   if [ -v appConf ]; then
      echo                     "Restore appconf ownership to : ${sUser}"
      echo "${Time} RootPKG   : Restore appconf ownership to : ${sUser}"            >> "$ERROR"

      find "${spkConf}" -print0 | while IFS= read -r -d '' dev; do
         chown "${sUser}" "${dev}"
         echo                     "Done : chown ${sUser} ${dev}"
         echo "${Time} RootPKG   : Done : chown ${sUser} ${dev}"                    >> "$ERROR"
      done
   fi

   # --------------------------------------------------------------------------------
   if [ -f "${sPackage}/startFailed" ]; then
      echo                     "Renaming : ${sPackage}/startFailed"
      echo "${Time} RootPKG   : Renaming : ${sPackage}/startFailed"                 >> "$ERROR"
      echo                     "To       : ${sPackage}/enabled"
      echo "${Time} RootPKG   : To       : ${sPackage}/enabled"                     >> "$ERROR"
      mv "${sPackage}/startFailed" "${sPackage}/enabled"
   fi

   # --------------------------------------------------------------------------------
   echo "---------------------------------------------------------------------------------"  >> "$ERROR"
   echo                     "${Name} \"successfully rooted\"."
   echo "${Time} RootPKG   : ${Name} \"successfully rooted\"."                               >> "$ERROR"
   echo "---------------------------------------------------------------------------------"  >> "$ERROR"
}
