# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------------#
# AvrLogger - Monitoring and visualization tool.                                     #
# Copyright (C) 2018 by BigRonin.                                                    #
# Member of the German Synology Community Forum "https://www.synology-forum.de"      #
# coding: utf-8                                                                      #
# -----------------------------------------------------------------------------------#
# This file is part of "AvrLogger.spk"                                               #
#                                                                                    #
# AvrLogger is free software: you can redistribute it and/or modify                  #
# it under the terms of the GNU General Public License as published by               #
# the Free Software Foundation, either version 3 of the License, or                  #
# (at your option) any later version.                                                #
# -----------------------------------------------------------------------------------#
import os
import sys
import glob

import asset
import basics   as io
import lib

# -----------------------------------------------------------------------------------
# --- Global options ...
# -----------------------------------------------------------------------------------
sUsbRootPath   = '/sys/bus/usb/devices/'

iUsbBasics     = [
                     'idVendor',
                     'idProduct',
                     'manufacturer',
                     'product',
                     'serial',
                     'version',
                     'speed',
                     'bMaxPower',
                 ]

# -----------------------------------------------------------------------------------
# --- Class to handle 'usb.ids' : 
# -----------------------------------------------------------------------------------
class ParseUsbIDs :

   # --------------------------------------------------------------------------------
   def __init__( self ) :
      self.UsbVendors  = dict()
      self.UsbProducts = dict()
      self.UsbClasses  = dict()
      self.iUsbLines   = []

   # --------------------------------------------------------------------------------
   # --- Return USB Vendor
   def FindUsbVendor( self, vID ) :
      if not self.UsbVendors :
         self._ParseVendor()

      return self.UsbVendors.get( vID.upper(), False )

   # --------------------------------------------------------------------------------
   # --- Return USB Product by Vendor
   def FindUsbProduct( self, vID, pID ) :
      if not self.UsbVendors :
         self._ParseVendor()

      return self.UsbProducts.get( ( vID.upper(), pID.upper() ), False )

   # --------------------------------------------------------------------------------
   # --- Return USB protocol from usbclasses list ...
   def FindUsbClass( self, cID, sID, pID ) :
      if not self.UsbClasses :
         self._ParseClasses()

      cID = cID.upper()
      sID = sID.upper()
      pID = pID.upper()

      sRet = self.UsbClasses.get( ( cID,  sID,  pID ) ) \
         or  self.UsbClasses.get( ( cID,  sID, 'FF' ) ) \
         or  self.UsbClasses.get( ( cID, 'FF', 'FF' ) )

      if sRet : return sRet
      else    : return self.UsbClasses[ 'FF', 'FF', 'FF' ]

   # --------------------------------------------------------------------------------
   # --- return True if all digits are valid hex digits ...
   def isHexDigit( self, sDigits ) :
      for dg in sDigits :
         if not dg.isdigit() and not dg in 'aAbBcCdDeEfF' : return False

      return True

   # --------------------------------------------------------------------------------
   # --- Parse /usr/share/usb.ids and fill UsbVendors, UsbProducts ...
   def _ParseVendor( self ) :
      if not self.iUsbLines :
         self.iUsbLines = io.ioRead( asset.avrUsbIdsFiles ).strip().split( '\n' )

      # -----------------------------------------------------------------------------
      # --- Parse usb.ids line by line ...
      for sLine in self.iUsbLines :
         sLine = sLine.rstrip()

         # --- Ignore blank or comment lines ...
         if not sLine or sLine[0] == '#' :
            continue

         # --------------------------------------------------------------------------
         if self.isHexDigit( sLine[0:4].upper() ) :
            vID = sLine[:4].upper()
            self.UsbVendors[ vID ] = sLine[6:].strip()
            continue

         # -----------------------------------------------------------------------------
         if sLine[0] == '\t' and self.isHexDigit( sLine[1:5] ) :
            pID = sLine[1:5].strip().upper()
            self.UsbProducts[ vID, pID ] = sLine[7:].strip()
            continue

         # -----------------------------------------------------------------------------
         break

   # --------------------------------------------------------------------------------
   # --- Parse /usr/share/usb.ids and fill UsbClasses ...
   def _ParseClasses( self ) :
      if not self.iUsbLines :
         self.iUsbLines = io.ioRead( asset.avrUsbIdsFiles ).strip().split( '\n' )

      f_Classes = False

      # -----------------------------------------------------------------------------
      # --- Parse usb.ids line by line ...
      for sLine in self.iUsbLines :
         sLine = sLine.rstrip()

         # --- Ignore blank or comment lines ...
         if not sLine or sLine[0] == '#' :
            continue

         # --------------------------------------------------------------------------
         if sLine[0] == 'C' :
            f_Classes = True
            cID   = sLine[2:4].strip().upper()
            cName = sLine[6:].strip()
            self.UsbClasses[ cID, 'FF', 'FF' ] = cName
            continue

         # --------------------------------------------------------------------------
         if not f_Classes :
            continue

         # --------------------------------------------------------------------------
         if sLine[0] == '\t' and self.isHexDigit( sLine[1:3] ) :
            dID = sLine[1:5].strip().upper()

            sName = sLine[5:].strip()
            if sName != 'Unused' and sName != '?' :
               sName = cName + ':' + sName
            else :
               sName = cName + ':'

            self.UsbClasses[ cID, dID, 'FF' ] = sName
            continue

         # --------------------------------------------------------------------------
         if sLine[0] == '\t' and sLine[1] == '\t' and self.isHexDigit( sLine[2:4] ) :
            pID = sLine[2:4].strip().upper()
            self.UsbClasses[ cID, dID, pID ] = sLine[6:].strip()
            continue

         # -----------------------------------------------------------------------------
         self.UsbClasses[ 'FF', 'FF', 'FF' ] = 'Vendor Specific'
         break

# -----------------------------------------------------------------------------------
# --- 
# -----------------------------------------------------------------------------------
def getUsbDevices ( sUsb=False, iNice=False, iDebug=False ) :

   sRet    = dict()
   Parse   = ParseUsbIDs()
   iHosts  = glob.glob( sUsbRootPath + '*:*/host*/target*/*:*/block/sd*'  )
   iHosts += glob.glob( sUsbRootPath + '*:*/host*/target*/*:*/block/usb*' )

#  for sPath in glob.glob( sUsbRootPath + '*:*/host*/target*/*:*/block/sd*' ) :
   for sPath in iHosts :
      sDev  = sPath.split( '/' )[-1]
      if sUsb and sDev != sUsb : continue

      if iDebug : print ( f'sUsbRootPath : {sPath}' )

      sPath = sPath.split( ':', 1 )[0]
      if iDebug : print ( f'  sPath : {sPath}' )

      for sAttr in iUsbBasics :
         try    : sRet[ sDev ].update( { sAttr : getAttr( sPath, sAttr ) } )
         except : sRet[ sDev ] =       { sAttr : getAttr( sPath, sAttr ) }
         if iDebug : print ( f'     sAttr : {sAttr:15} = {getAttr( sPath, sAttr )}' )

      # -----------------------------------------------------------------------------
      if not sRet[ sDev ][ 'manufacturer' ] or not sRet[ sDev ][ 'product' ] :
         sRet[ sDev ][ 'manufacturer' ] = Parse.FindUsbVendor ( sRet[ sDev ][ 'idVendor' ] )
         sRet[ sDev ][ 'product' ]      = Parse.FindUsbProduct( sRet[ sDev ][ 'idVendor' ], sRet[ sDev ][ 'idProduct' ] )

      # -----------------------------------------------------------------------------
      # --- Some USB Card readers have a better name then Generic ...

      sTemp = Parse.FindUsbVendor( sRet[ sDev ][ 'idVendor' ] )
      if sTemp : sRet[ sDev ][ 'manufacturer' ] = sTemp

      sTemp = Parse.FindUsbProduct( sRet[ sDev ][ 'idVendor' ], sRet[ sDev ][ 'idProduct' ] )
      if sTemp : sRet[ sDev ][ 'product' ] = Parse.FindUsbProduct( sRet[ sDev ][ 'idVendor' ], sRet[ sDev ][ 'idProduct' ] )

      # -----------------------------------------------------------------------------
      if iNice :
         cFormat = io.ExtendedFormatter()

         sRet[ sDev ]['UsbID'    ] = cFormat.format( '{idVendor!u}:{idProduct!u}', **sRet[ sDev ] )
         sRet[ sDev ]['UsbName'  ] = '{manufacturer} : {product}'.format( **sRet[ sDev ] )
         sRet[ sDev ]['UsbSerial'] = sRet[ sDev ]['serial']
         sRet[ sDev ]['UsbSpeed' ] = '[USB {version}, {speed:>4.4} Mbps, {bMaxPower:>5.5}]'.format( **sRet[ sDev ] )
         sRet[ sDev ]['UsbMount' ] = False
         sRet[ sDev ]['UsbShare' ] = False

         # --------------------------------------------------------------------------
         # --- Get Mount-Point for a USB-Device by Device-Name ...
         sShare = []
         sMount = []
         sTemp = [ x for x in lib.getCMD( '/usr/bin/mount | grep /dev/' + sDev ).strip().split( '\n' ) if x ]

         if sTemp :
            for sLine in sTemp :
               sLine = sLine.split( 'on ',   1 )[-1]
               sLine = sLine.split( ' type', 1 )[0]
               if iDebug : print ( f'        Mount-Point : {sLine}' )

               # --------------------------------------------------------------------
               # --- Add Mount-Point
               sMount.append( sLine )
               # --------------------------------------------------------------------
               # --- Add Share-Name
               sShare.append( lib.getCMD( 'synoshare --enummnt ' + sLine ).strip().split( '\n' )[1] )
               if iDebug : print ( f'        Share-Name  : ' + lib.getCMD( 'synoshare --enummnt ' + sLine ).strip().split( '\n' )[1] )


         if sTemp :
            sRet[ sDev ]['UsbMount'] = sMount
            sRet[ sDev ]['UsbShare'] = sShare
         else :
            sRet[ sDev ]['UsbMount'] = [ 'Device is not mounted' ]
            sRet[ sDev ]['UsbShare'] = [ 'Device has no shares'  ]

   # --------------------------------------------------------------------------------
   return sRet

# -----------------------------------------------------------------------------------
# --- Return a Attribut :
# -----------------------------------------------------------------------------------
def getAttr ( sPath, sName ) :

   return io.ioRead( sPath + '/' + sName )

   sAttr = sPath + '/' + sName

   try :
      with open( sAttr, 'r', encoding='utf-8' ) as f:
         return ' '.join( f.read().split() )
   except :
      return False

# --------------------------------------- END ---------------------------------------
# -----------------------------------------------------------------------------------
#  sys.exit()
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
